/* Author: Alain Thellier - Paris - 2010 . See ReadMe for more infos								*/
/* v3.0 Sept  2011: Draw FPS correctly on Aros too. Now works with Wazp3D using "hard overlay"				*/
/* v4.0 March 2012: Try to use W3D_InterleavedArray() on OS4 									*/
/* v5.0 Sept  2012: Use W3D_DrawElement() on OS4 ==> faster										*/
/* v6.0 Feb   2016: Use Nova on OS4															*/
/* v6.V March 2017: Vampire test															*/

#define DISPLAYW 640
#define DISPLAYH 480
#define COWSIZE 90.0



#ifdef __amigaos4__
#define OS4
#else
#define OS3
#endif

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <strings.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <proto/cybergraphics.h>


#if !defined(PIXFMT_LUT8)
#include "cybergraphics.h"
#endif


/*==================================================================*/
#include "Cow3D_Object.h"				/* 3d object*/
UBYTE  progname[]={"CoW3D6V/OS3: type 'b' for speed up"};
/*==================================================================*/
#define LIBCLOSE(libbase)	 if(libbase!=NULL)	{CloseLibrary( (struct Library  *)libbase );   libbase=NULL; }
#define LIBOPEN(libbase,name,version)  libbase	=(void*)OpenLibrary(#name,version);				if(libbase==NULL) 	return(FALSE);
/*==================================================================*/
#define LL            {if(debug) printf("Line:%ld\n",(ULONG)__LINE__);}
#define VAR(var)   if(debug) {printf(" " #var "=" ); printf("%ld\n",  ((ULONG)var)  );}
#define   VARS(var)   if(debug) {printf(" " #var "=<%s>;\n",var); }
#define ZZ  if(debug) printf("ZZ stepping ..\n");
#define REM(message)   if(debug) printf(#message"\n");
#define MYCLR(x) 	memset(&x,0,sizeof(x));
#define NLOOP(nbre) for(n=0;n<nbre;n++)
#define MLOOP(nbre) for(m=0;m<nbre;m++)
#define SPEED 5  /* object speed */
#define AA            { asmline= __LINE__ ;  debugprintf("Line:%ld\n",asmline); asmline= 0x1234;  }
#define LLL				PrintLine(C,#__FUNCTION__,__LINE__); 
#define FUNC            if(debug) {printf("%s\n",__FUNCTION__);}
#define FUNCOK          if(debug) {printf("%s:OK\n",__FUNCTION__);}
#define FRAMESCOUNT (360/SPEED)		/* frames count to rotate fully */
#define MAXPRIM (3*1000)			/* Wazp3D cant draw bigger primitives  so split the object in several drawing*/
BOOL debug=FALSE;

/*==================================================================*/
struct Point3D{
	float x,y,z;
	float u,v,w;
	UBYTE RGBA[4];
};
/*==================================================================*/
struct Texture3D{
	UBYTE *pixels;				/* texture data */
	UWORD w,h,bpp,pad;
};
/*==================================================================*/
struct Object3D{
	struct Point3D* P;				/* original points */
	struct Point3D* P2;				/* points rotated/transformed to screen */
	ULONG *PI;				/* points' indices that define the triangles */
	ULONG Pnb;				/* points counts(french "points nombre" */
	ULONG PInb;				/* PI counts(french "PI nombre" */
	struct Texture3D  *T;			/* our texture */
};
/*==================================================================*/
struct GfxBase*			GfxBase				=NULL;
struct IntuitionBase*	IntuitionBase		=NULL;
struct Library*			CyberGfxBase		=NULL;
/*==================================================================*/
struct Cow3D{
struct Screen  *screen;
struct Window  *window;
struct BitMap  *bufferbm;
APTR Ctx;
ULONG result,NumTMU,ModeID,flags,ScreenBits;
BOOL  closed;
BOOL greyreadz;
BOOL colored;
BOOL UseGouraud;
BOOL showfps;
BOOL Zbuffer;
BOOL zupdate;
BOOL HideFace;
BOOL rotate;
BOOL DrawCow;
BOOL DrawCosmos;
UBYTE drawmode;
BOOL bigpoint;
BOOL UseLineTest;
BOOL UseZTest;
BOOL IsBuffered;
ULONG Zmode;
BOOL optimroty;
float ViewMatrix[16];
float RotY;
ULONG FramesCounted;
ULONG FPS;
UBYTE FpsText[200];
UBYTE zname[5];
struct Object3D *CowObj;
struct Object3D *CosmosObj;
struct Object3D *QuadObj;
struct Point3D  *BufferedP;			/* store all transformed points */
UBYTE Pdone[FRAMESCOUNT];
ULONG SrcFunc,DstFunc,SrcDstFunc;
ULONG TexEnvMode;
ULONG DisplayW,DisplayH;
ULONG ScreenW,ScreenH;

BOOL FrameLimit;
ULONG Time, OldTime;

float opaqueBlack[4];
double clearDepth;

};
struct Cow3D C;

#include "Vampire3D.c"				/* raw triangle renderer */

/*==================================================================*/
void SetDefault()
{

/* all other values are set to zero */	
	MYCLR(C);

	C.DisplayW=DISPLAYW;
	C.DisplayH=DISPLAYH;
	C.showfps=TRUE;
	C.rotate=TRUE;
	C.DrawCow=TRUE;
	C.optimroty=TRUE;
	
	C.opaqueBlack[0] = 0;
	C.opaqueBlack[1] = 0;
	C.opaqueBlack[2] = 64;
	C.opaqueBlack[3] = 255;
	C.clearDepth = 1.0;
		
}
/*==================================================================*/
void SetMrot(float *M,float R ,UBYTE a,UBYTE b,UBYTE c,UBYTE d);
void SetMrx(float *M,float x);
void SetMry(float *M,float y);
void SetMrz(float *M,float z);
void TransformP(register float *M,struct Point3D  *P,LONG Pnb);
void YrotateP(register float *M,struct Point3D  *P,LONG Pnb);
void ProjectP(struct Point3D  *P,LONG Pnb);
/*==================================================================*/
BOOL DoViewWarp3D(void)
{
ULONG frame,n;
BOOL framebuffered;		
struct Point3D *P;
	
		framebuffered=FALSE;

		if(C.IsBuffered)
		if(C.BufferedP==NULL)
			{
			printf("Will allocate buffer %ld bytes\n",FRAMESCOUNT*C.CowObj->Pnb*sizeof(struct Point3D));
			C.BufferedP = malloc(FRAMESCOUNT*C.CowObj->Pnb*sizeof(struct Point3D));
			if(!C.BufferedP) printf("no memory for buffer!\n");
			NLOOP(FRAMESCOUNT) C.Pdone[n]=FALSE;
			}


		if(C.rotate)
		{
		frame=C.RotY/SPEED;

		if(C.IsBuffered)
		if(C.BufferedP!=NULL)
		if(C.Pdone[frame])		/* use a saved frame ? */
			{
			C.CowObj->P2		=&C.BufferedP[C.CowObj->Pnb*frame];	/* P2=IsBuffered points */
			framebuffered=TRUE;
			}

		if(!framebuffered)		/* then transform/rotate/project frame  */
			{
			P=C.CowObj->P;
			C.CowObj->P2		=&P[C.CowObj->Pnb];		/* use default P2 */
			memcpy(C.CowObj->P2,C.CowObj->P,C.CowObj->Pnb*sizeof(struct Point3D));	/* copy the object's points to P2*/
			SetMry(C.ViewMatrix,C.RotY);	/* do a y rotation matrix */
			if(C.optimroty)
				YrotateP(C.ViewMatrix,C.CowObj->P2,C.CowObj->Pnb);		/* only y-rotate the object's points in P2*/
			else
				TransformP(C.ViewMatrix,C.CowObj->P2,C.CowObj->Pnb);		/* fully transform the object's points in P2*/
			ProjectP(C.CowObj->P2,C.CowObj->Pnb);/* project points to screen */
			}

		if(C.IsBuffered)
		if(C.BufferedP!=NULL)
		if(!C.Pdone[frame])		/* then save this frame */
			{
			memcpy(&C.BufferedP[C.CowObj->Pnb*frame],C.CowObj->P2,C.CowObj->Pnb*sizeof(struct Point3D));	/* save the transformed points to buffer */
			C.Pdone[frame]=TRUE;
			}

		}
	
	return(TRUE);
}
/*=================================================================*/
void SetP(struct Point3D  *P,float x, float y, float z,float w,float u, float v,UBYTE r,UBYTE g,UBYTE b,UBYTE a)
{
	if(P==NULL) return;

	P->x=x; 
	P->y=y;
	P->z=z;
	
	P->u=u;
	P->v=v;
	P->w=w;

	P->RGBA[0]=r;
	P->RGBA[1]=g;
	P->RGBA[2]=b;
	P->RGBA[3]=a;
}
/*==================================================================*/
void DrawObject(struct Object3D *O)
{
	BindTextureV(C.Ctx,O->T);			
	SetArrayV(C.Ctx,O->P2,O->PI,O->PInb);
	DrawTrianglesV(C.Ctx);
}	
/*=================================================================*/
struct Object3D *AddObject(ULONG Pnb,ULONG PInb)
{
struct Object3D *O;
ULONG Osize,Psize,P2size,PIsize,size;
UBYTE *pt;

	Osize =1   *sizeof(struct Object3D);	
	Psize =Pnb *sizeof(struct Point3D);
	P2size=Pnb *sizeof(struct Point3D);
	PIsize=PInb*sizeof(ULONG);

	size=Osize+PIsize+Psize+P2size; /* define all the object datas in a single malloc */
	pt=malloc(size);

	memset(pt,0,sizeof(struct Object3D));
	
	O			=(struct Object3D *)pt;		pt=pt+Osize;
	O->PI		=(ULONG *)pt;				pt=pt+PIsize;
	O->P		=(APTR)pt;					pt=pt+Psize;	
	O->P2		=(APTR)pt;					pt=pt+P2size;
	O->Pnb		=Pnb;
	O->PInb		=PInb;

	return(O);
}
/*=================================================================*/
void ReadIncludedObjectWarp3D(struct Object3D *O,float *V,ULONG *i)
{
struct Point3D *P    = NULL;
float x,y,z,w,u,v;
UBYTE r,g,b,a;
ULONG n,Psize;


	NLOOP(O->PInb)
		O->PI[n]=i[n];

	r=g=b=a=255;

	P=O->P;
	NLOOP(O->Pnb)
	{
	u=V[0]; v=V[1]; x=V[2]; y=V[3]; y=-y; z=V[4]; w=1.0/z;	/* beware z is negated so faces are reverted again in good order*/
	r=x*90+90; g=y*90+90; b=z*90+90;	/* just set various colors a stupid way*/
	SetP(P,x,y,z,w,u,v,r,g,b,a); 
	P++; 
	V=V+5;
	}

/* default: P2 are untransformed P */

	Psize=O->Pnb *sizeof(struct Point3D);
	memcpy(O->P2,O->P,Psize);
}
/*=================================================================*/
void SetMrot(float *M,float R ,UBYTE a,UBYTE b,UBYTE c,UBYTE d)
{	/* define a rotation matrix */
float Pi =3.1416;
WORD n;
	if(M==NULL) return;
	M[0]=M[5]=M[10]=M[15]= 1.0;  	M[1]=M[2]=M[3]=M[4]=M[6]=M[7]=M[8]=M[9]=M[11]=M[12]=M[13]=M[14]=0.0;
	if(R==0.0) return;
	if(R>=360.0)
		{n=R/360.0;  R=R-n*360.0;}
	if(R<0.0)
		{n=1+R/360.0;R=R-n*360.0;}
	R= R / 180.0 * Pi;
	M[a]  =  (float)cos(R);
	M[b]  =  M[a];
	M[c]  =  (float)sin(R);
	M[d]  = -M[c];
}
/*=================================================================*/
void SetMrx(float *M,float x)		/* define X rotation matrix */
	{ SetMrot(M,x,5,10,6,9); }
/*=================================================================*/
void SetMry(float *M,float y)		/* define Y rotation matrix */
	{ SetMrot(M,y,10,0,8,2); }
/*=================================================================*/
void SetMrz(float *M,float z)		/* define Z rotation matrix */
	{ SetMrot(M,z,0,5,1,4); }
/*=================================================================*/
void TransformP(register float *M,struct Point3D  *P,LONG Pnb)
{	/* transform points with a given matrix */
register float x;
register float y;
register float z;
register LONG n;

NLOOP(Pnb)
	{
	x=P->x;y=P->y;z=P->z;
	P->x= M[0]*x + M[4]*y+ M[8] *z+ M[12];
	P->y= M[1]*x + M[5]*y+ M[9] *z+ M[13];
	P->z= M[2]*x + M[6]*y+ M[10]*z+ M[14];
	P++;
	}
}
/*=================================================================*/
void YrotateP(register float *M,struct Point3D  *P,LONG Pnb)
{	/* y rotate points with a given matrix : optimized TransformP() */
register float x;
register float z;
register float m0 =M[0];
register float m8 =M[8];
register float m2 =M[2];
register float m10=M[10];
register LONG n;

NLOOP(Pnb)
	{
	x=P->x;z=P->z;
	P->x= m0*x +  m8 *z;
	P->z= m2*x +  m10*z;
	P++;
	}
}
/*=================================================================*/
void ProjectP(struct Point3D  *P,LONG Pnb)
{	/* project points to screen-coordinates */
register float x;
register float y;
register float z;
register float sizex;
register float sizey;
register float sizez;
register LONG n;

	sizex=C.DisplayW/2.0;
	sizey=C.DisplayH /2.0;
	sizez=0.8  /2.0;
NLOOP(Pnb)
	{
	x=P->x;y=P->y;z=P->z;
	P->x= x*sizex+sizex;
	P->y= y*sizey+sizey;
	P->z= z*sizez+sizez;
	P->w= 1.0/P->z;
	P++;
	}
}
/*=================================================================*/
BOOL LoadFile(UBYTE *buffer, UBYTE *name,ULONG bufferSize)
{	/* load a file in memory */
FILE *fp;
ULONG size=0;

	fp = fopen(name,"rb");
	if(fp == NULL)
		{printf("Cant open file ! <%s>\n",name); return(FALSE); }
	size = fread(buffer,bufferSize,1,fp);
	if(size ==0 )
		{printf("Cant read file ! <%s>\n",name); return(FALSE); }
	fclose(fp);
	return(TRUE);
}
/*=================================================================*/
BOOL LoadTexture(struct Object3D *O,UBYTE* name,UWORD size,UWORD bits)
{	/* load a RAW rgb bitmap in memory. Make it a Warp3D texture */
ULONG texsize = size*size*bits/8;

	O->T = (struct Texture3D *) malloc(sizeof(struct Texture3D));
	O->T->pixels = (UBYTE *) 	malloc(texsize);
	if(O->T->pixels==NULL)
		{printf("Cant alloc picture! \n");return(FALSE);}
	if(LoadFile(O->T->pixels,name,texsize)==FALSE)
		{printf("Cant load picture! \n");return(FALSE);}
	O->T->w=size;		
	O->T->h=size;		
	O->T->bpp=32/8;		
	return(TRUE);	
}
/*==========================================================================*/
void	 MyDrawText(WORD x,WORD y,UBYTE *text)
{	/* draw a text in the window */
struct RastPort *rp;


	rp=C.window->RPort;
	SetAPen(rp, 0) ;
	RectFill(rp,x-3,y-9,x+8*strlen(text)+3,y+2);

	SetAPen(rp, 1);
	Move(rp,x-1,y-1);
	Text(rp,(void*)text, strlen(text));

	SetAPen(rp, 1);
}
/*=================================================================*/
static void DoFps(void)
{/* count the average "frame per second" */


	if (++C.FramesCounted >= (20*FRAMESCOUNT)) 			/* ie after the object turned 20 times  */
	{
		C.Time  = clock();
		C.FPS = (20*FRAMESCOUNT*CLOCKS_PER_SEC) / (C.Time - C.OldTime) ;
		C.OldTime = C.Time; 
		C.FramesCounted = 0;
	}

	sprintf(C.FpsText,"COW3D: %ld Fps(%ldtris %ldpoints) optirot%d buffer%d)",C.FPS,C.CowObj->PInb/3,C.CowObj->Pnb,C.optimroty,C.IsBuffered);	
	
	if(C.FramesCounted==0)
	if(C.Time!=0)
		printf("%s\n",C.FpsText);	

}
/*==================================================================================*/
BOOL OpenAmigaLibraries(void)
{
	LIBOPEN(GfxBase,graphics.library,0L)
	LIBOPEN(IntuitionBase,intuition.library,0L)
	LIBOPEN(CyberGfxBase,cybergraphics.library,0L)
	
	return(TRUE);
}
/*======================================================================================*/
void CloseAmigaLibraries()
{
	LIBCLOSE(GfxBase)
	LIBCLOSE(IntuitionBase)
	LIBCLOSE(CyberGfxBase)
}
/*==========================================================================*/
BOOL OpenDisplay(void)
{	/* open a window & a rastport ("back buffer") & open warp3d & create a warp3d C.Ctx  */
ULONG Flags =WFLG_ACTIVATE | WFLG_REPORTMOUSE | WFLG_RMBTRAP | WFLG_SIMPLE_REFRESH | WFLG_GIMMEZEROZERO ;
ULONG IDCMPs=IDCMP_CLOSEWINDOW | IDCMP_VANILLAKEY | IDCMP_RAWKEY | IDCMP_MOUSEMOVE | IDCMP_MOUSEBUTTONS ;
	
	if(OpenAmigaLibraries()==FALSE)
		return(FALSE);

	REM(OpenDisplay)
	C.screen 	=LockPubScreen("Workbench") ;
	C.ScreenW	=C.screen->Width;
	C.ScreenH	=C.screen->Height;
	C.ModeID  =GetVPModeID(&C.screen->ViewPort);
	UnlockPubScreen(NULL, C.screen);

	C.window = OpenWindowTags(NULL,
	WA_Activate,		TRUE,
	WA_InnerWidth,		C.DisplayW,
	WA_InnerHeight,		C.DisplayH,
	WA_Left,			(C.ScreenW - C.DisplayW)/2,
	WA_Top,				(C.ScreenH - C.DisplayH)/2,
	WA_Title,			(ULONG)progname,
	WA_DragBar,			TRUE,
	WA_CloseGadget,		TRUE,
	WA_GimmeZeroZero,	TRUE,
	WA_Backdrop,		FALSE,
	WA_Borderless,		FALSE,
	WA_IDCMP,			IDCMPs,
	WA_Flags,			Flags,
	WA_Flags,			Flags,
	TAG_DONE);

	if(C.window==NULL)
		{printf("Cant open window\n");return FALSE;}

	C.Ctx=CreateContextV(C.window,C.DisplayW,C.DisplayH);
	if(C.Ctx==NULL)
		{printf("Cant open Ctx\n");return FALSE;}
		
	return(TRUE);
}
/*=================================================================*/
void FreeObj(struct Object3D *O)
{
	FUNC
	if(O->T)
	if(O->T->pixels)
		{free(O->T->pixels); O->T->pixels=NULL;}
		
	if(O->T)
		{free(O->T); O->T=NULL;}
		
	free(O);	/* free also P P2 PI that are included in Object */
}
/*=================================================================*/
void FreeData(void)
{

REM(BufferedP----------)
	if(C.BufferedP)		free(C.BufferedP);
REM(CowObj----------)
	if(C.CowObj)		FreeObj(C.CowObj);
	if(C.CosmosObj)		FreeObj(C.CosmosObj);
}
/*==========================================================================*/
void CloseDisplay(void)
{
REM(CloseDisplay);	
	
	DeleteContextV(C.Ctx);

	if(C.window)					CloseWindow(C.window);

	CloseAmigaLibraries();
}	
/*================================================================================*/
void WindowEvents(void)
{		/* manage the window  */
struct IntuiMessage *imsg;

REM(WindowEvents)
	while( (imsg = (struct IntuiMessage *)GetMsg(C.window->UserPort)))
	{
	if(imsg == NULL) break;
	switch (imsg->Class)
		{
		case IDCMP_CLOSEWINDOW:
			C.closed=TRUE;				break;
		case IDCMP_VANILLAKEY:
			switch(imsg->Code)
			{
			case 'f':	C.showfps=!C.showfps;		break;
			case 'o':	C.optimroty=!C.optimroty;	break;
			case 'b':	C.IsBuffered=!C.IsBuffered;	break;
			case '!':	debug=!debug;	break;
			case 'F':	C.FrameLimit++;	if(C.FrameLimit>10) C.FrameLimit=0; break;
			case 27:	C.closed=TRUE;		break;
			default:break;
			}
			break;
		default:
			break;
		}
	if(imsg)
		{ReplyMsg((struct Message *)imsg);imsg = NULL;}
	}
}
/*==========================================================================*/		
void UpdateDisplay(struct Context3D* C)
{
	WritePixelArray(C->pixels,0,0,C->w*32/8,C->window->RPort,0,0,C->w,C->h,RECTFMT_RGBA);
//	WritePixelArray(C->zbuffer,0,0,C->w*32/8,C->window->RPort,0,0,C->w,C->h,RECTFMT_RGBA);
}	
/*=================================================================*/
int main(int argc, char *argv[])
{
	printf("Cow3D6V: build " __DATE__ " at " __TIME__ "\n");
	SetDefault();
			
	if(!OpenDisplay())
		goto panic;
	
	
	C.CowObj =AddObject(pointsCount,trianglesCount*3);				/* the cow object */
	ReadIncludedObjectWarp3D(C.CowObj,points,indices);
	LoadTexture(C.CowObj,TEXNAME,TEXSIZE,32);						/* texture data as R8 G8 B8 A8*/

	C.CosmosObj=AddObject(pointsCountQuad,trianglesCountQuad*3);	/* simple quad as 2 tris */
	ReadIncludedObjectWarp3D(C.CosmosObj,pointsQuad,indicesQuad);
	LoadTexture(C.CosmosObj,"Cosmos_256X256X32.RAW",256,32);		/* texture data as R8 G8 B8 A8*/
	
	while(!C.closed)
	{
	REM(====== New frame =====)

		DoViewWarp3D();		
		
		C.UseGouraud=TRUE;
		DrawObject(C.CowObj);		/* draw it */
		
		C.UseGouraud=FALSE;
		DrawObject(C.CosmosObj);	/* draw it */
		
		UpdateV(C.Ctx,C.window);	/* copy to window */

		WindowEvents();					/* is window closed ? */
		if(C.showfps) DoFps(); 			/* count the average "frames per second" */
		if(C.rotate)
			C.RotY=C.RotY+SPEED;			/* rotate */
		if(C.RotY>=360.0)
				C.RotY=0.0;				/* rotate */
	}
	


panic:
	REM(ending...)
	FreeData();
	CloseDisplay();
	return 0;
}
/*==================================================================*/








			
			

				











